from ._config import (
    BaseMetricsConfiguration,
    DisabledMetricsConfiguration,
    EventsConfiguration,
    KafkaMetricsConfiguration,
    MetricsConfiguration,
    MockMetricsConfiguration,
    metrics_configuration_factory,
)
from ._event_manager import (
    EventManager,
    EventPublisher,
    FailedEventPublisher,
    KafkaEventManager,
    KafkaEventPublisher,
    MockEventManager,
    MockEventPublisher,
    NoopEventManager,
    NoopEventPublisher,
)
from ._exceptions import (
    DuplicateEventError,
    EventManagerUnintializedError,
    EventManagerUsageError,
    KafkaTopicError,
    UnsupportedAvroSchemaError,
)
from ._models import EventMetadata, EventPayload
from ._testing import (
    ANY,
    NOT_NONE,
    BaseAssertionError,
    NotPublishedConsecutivelyError,
    NotPublishedError,
    PublishedCountError,
    PublishedList,
    PublishedTooFewError,
)

__all__ = [
    "ANY",
    "NOT_NONE",
    "BaseAssertionError",
    "BaseMetricsConfiguration",
    "DisabledMetricsConfiguration",
    "DuplicateEventError",
    "EventManager",
    "EventManagerUnintializedError",
    "EventManagerUsageError",
    "EventMetadata",
    "EventPayload",
    "EventPublisher",
    "EventsConfiguration",
    "FailedEventPublisher",
    "KafkaEventManager",
    "KafkaEventPublisher",
    "KafkaMetricsConfiguration",
    "KafkaTopicError",
    "MetricsConfiguration",
    "MockEventManager",
    "MockEventPublisher",
    "MockMetricsConfiguration",
    "NoopEventManager",
    "NoopEventPublisher",
    "NotPublishedConsecutivelyError",
    "NotPublishedError",
    "PublishedCountError",
    "PublishedList",
    "PublishedTooFewError",
    "UnsupportedAvroSchemaError",
    "metrics_configuration_factory",
]



"""Mock Google Cloud Storage API for testing."""

from __future__ import annotations

from collections.abc import Iterator
from datetime import UTC, datetime, timedelta
from io import BufferedReader
from pathlib import Path
from typing import Any
from unittest.mock import Mock, patch

try:
    from google.cloud import storage
except ImportError as e:
    raise ImportError(
        "The safir.testing.gcs module requires the gcs extra. "
        "Install it with `pip install safir[gcs]`."
    ) from e

__all__ = [
    "MockBlob",
    "MockBucket",
    "MockStorageClient",
    "patch_google_storage",
]


class MockBlob(Mock):
    """Mock version of ``google.cloud.storage.blob.Blob``.

    Parameters
    ----------
    name
        Name of the blob.
    expected_expiration
        The expiration that should be requested in a call to
        ``generate_signed_url`` on an underlying blob.  A non-matching call
        will produce an assertion failure.
    """

    def __init__(
        self, name: str, expected_expiration: timedelta | None = None
    ) -> None:
        super().__init__(spec=storage.blob.Blob)
        self.name = name
        self._expected_expiration = expected_expiration

    def generate_signed_url(
        self,
        *,
        version: str,
        expiration: timedelta,
        method: str,
        response_type: str | None = None,
        credentials: Any | None = None,
    ) -> str:
        """Generate a mock signed URL for testing.

        Parameters
        ----------
        version
            Must be ``v4``.
        expiration
            Must match the ``expected_expiration`` argument to the
            constructor if it was given.
        method
            Must be ``GET``.
        response_type
            May be anything and is ignored.
        credentials
            May be anything and is ignored.

        Returns
        -------
        str
            Always returns :samp:`https://example.com/{name}` where *name* is
            the name of the blob.
        """
        assert version == "v4"
        if self._expected_expiration:
            assert expiration == self._expected_expiration
        assert method == "GET"
        return f"https://example.com/{self.name}"


class MockFileBlob(MockBlob):
    """Mock version of ``google.cloud.storage.blob.Blob`` for a file.

    Parameters
    ----------
    name
        Name of the blob.
    path
        Path to the file for this blob.
    expected_expiration
        The expiration that should be requested in a call to
        ``generate_signed_url`` on an underlying blob.  A non-matching call
        will produce an assertion failure.

    Attributes
    ----------
    size : `int`
        Size of the underlying file.
    updated : `datetime.datetime`
        When the underlying file was last updated.
    etag : `str`
        Etag value for the file (taken from its inode number).
    """

    def __init__(
        self,
        name: str,
        path: Path,
        expected_expiration: timedelta | None = None,
    ) -> None:
        super().__init__(name, expected_expiration)
        self._path = path
        self._exists = path.exists()
        if self._exists:
            self.size = self._path.stat().st_size
            mtime = self._path.stat().st_mtime
            self.updated = datetime.fromtimestamp(mtime, tz=UTC)
            self.etag = str(self._path.stat().st_ino)

    def download_as_bytes(self) -> bytes:
        """Get contents of the blob.

        Returns
        -------
        bytes
            Contents of the underlying file.
        """
        return self._path.read_bytes()

    def exists(self) -> bool:
        """Whether the underlying file exists.

        Returns
        -------
        bool
            `True` if it does, `False` otherwise.
        """
        return self._exists

    def open(self, mode: str) -> BufferedReader:
        """Open the file.

        Parameters
        ----------
        mode
            Mode with which to open it (must be ``rb`` or an assertion failure
            is raised).

        Returns
        -------
        BufferedReader
            Stream representing the file.

        Raises
        ------
        AssertionFailure
            Unexpected mode argument.
        """
        assert mode == "rb"
        return self._path.open("rb")

    def reload(self) -> None:
        """Reload the metadata for the file.

        This does nothing in the mock.
        """


class MockBucket(Mock):
    """Mock version of ``google.cloud.storage.bucket.Bucket``.

    Parameters
    ----------
    expected_expiration
        The expiration that should be requested in a call to
        ``generate_signed_url`` on an underlying blob.  A non-matching call
        will produce an assertion failure.
    path
        Root of the file path for blobs, if given.  If not given, a simpler
        mock blob will be used that only supports ``generate_signed_url``.
    """

    def __init__(
        self,
        bucket_name: str,
        expected_expiration: timedelta | None = None,
        path: Path | None = None,
    ) -> None:
        super().__init__(spec=storage.bucket.Bucket)
        self._expected_expiration = expected_expiration
        self._path = path

    def blob(self, blob_name: str) -> MockBlob:
        """Retrieve a mock blob.

        Parameters
        ----------
        blob_name
            The name of the blob, used later to form its signed URL.

        Returns
        -------
        MockBlob
            The mock blob.
        """
        if self._path:
            return MockFileBlob(
                blob_name, self._path / blob_name, self._expected_expiration
            )
        else:
            return MockBlob(blob_name, self._expected_expiration)


class MockStorageClient(Mock):
    """Mock version of ``google.cloud.storage.Client``.

    Only supports `bucket`, and the resulting object only supports the
    ``blob`` method.  The resulting blob only supports the
    ``generate_signed_url`` method.

    Parameters
    ----------
    expected_expiration
        The expiration that should be requested in a call to
        ``generate_signed_url`` on an underlying blob.  A non-matching call
        will produce an assertion failure.
    path
        Root of the file path for blobs, if given.  If not given, a simpler
        mock blob will be used that only supports ``generate_signed_url``.
    bucket_name
        If set, all requests for a bucket with a name other than the one
        provided will produce assertion failures.
    """

    def __init__(
        self,
        expected_expiration: timedelta | None = None,
        path: Path | None = None,
        bucket_name: str | None = None,
    ) -> None:
        super().__init__(spec=storage.Client)
        self._bucket_name = bucket_name
        self._expected_expiration = expected_expiration
        self._path = path

    def bucket(self, bucket_name: str) -> MockBucket:
        """Retrieve a mock bucket.

        Parameters
        ----------
        bucket_name
            Name of the bucket.  If a bucket name was given to the
            constructor, this name will be checked against that one and a
            mismatch will cause an assertion failure.

        Returns
        -------
        MockBucket
            The mock bucket.
        """
        if self._bucket_name:
            assert bucket_name == self._bucket_name
        return MockBucket(bucket_name, self._expected_expiration, self._path)


def patch_google_storage(
    *,
    expected_expiration: timedelta | None = None,
    path: Path | None = None,
    bucket_name: str | None = None,
) -> Iterator[MockStorageClient]:
    """Replace the Google Cloud Storage API with a mock class.

    This function will replace the ``google.cloud.storage.Client`` API with a
    mock object.  It only supports bucket requests, the buckets only support
    blob requests, and the blobs only support requests for signed URLs.  The
    value of the signed URL will be :samp:`https://example.com/{blob}` where
    *blob* is the name of the blob.

    Parameters
    ----------
    expected_expiration
        The expiration that should be requested in a call to
        ``generate_signed_url`` on an underlying blob.  A non-matching call
        will produce an assertion failure.
    path
        Root of the file path for blobs, if given.  If not given, a simpler
        mock blob will be used that only supports ``generate_signed_url``.
    bucket_name
        If set, all requests for a bucket with a name other than the one
        provided will produce assertion failures.

    Yields
    ------
    MockStorageClient
        The mock Google Cloud Storage API client (although this is rarely
        needed by the caller).

    Notes
    -----
    This function also mocks out ``google.auth.default`` and the impersonated
    credentials structure so that this mock can be used with applications that
    use workload identity.

    To use this mock successfully, you must not import ``Client`` (or
    ``Credentials``) directly into the local namespace, or it will not be
    correctly patched.  Instead, use:

    .. code-block:: python

       from google.cloud import storage

    and then use ``storage.Client`` and so forth.  Do the same with
    q`google.auth.impersonated_credentials.Credentials``.

    Examples
    --------
    Normally this should be called from a fixture in ``tests/conftest.py``
    such as the following:

    .. code-block:: python

       from datetime import timedelta

       from safir.testing.gcs import MockStorageClient, patch_google_storage


       @pytest.fixture
       def mock_gcs() -> Iterator[MockStorageClient]:
           yield from patch_gcs(
               expected_expiration=timedelta(hours=1),
               bucket_name="some-bucket",
           )
    """
    mock_gcs = MockStorageClient(expected_expiration, path, bucket_name)
    with patch("google.auth.impersonated_credentials.Credentials"):
        with patch("google.auth.default", return_value=(None, None)):
            with patch("google.cloud.storage.Client", return_value=mock_gcs):
                yield mock_gcs



"""Service layer for a UWS service."""

from __future__ import annotations

import asyncio
from contextlib import suppress
from datetime import datetime, timedelta

from structlog.stdlib import BoundLogger

from safir.datetime import isodatetime

try:
    from vo_models.uws import Jobs, JobSummary
    from vo_models.uws.types import ExecutionPhase
    from vo_models.vosi.availability import Availability

    from safir.arq import ArqQueue, JobMetadata
    from safir.arq.uws import WorkerJobInfo
except ImportError as e:
    raise ImportError(
        "The safir.uws module requires the uws extra. "
        "Install it with `pip install safir[uws]`."
    ) from e

from ._config import UWSConfig
from ._constants import JOB_STOP_TIMEOUT
from ._exceptions import (
    InvalidPhaseError,
    SyncJobFailedError,
    SyncJobNoResultsError,
    SyncJobTimeoutError,
)
from ._models import (
    ACTIVE_PHASES,
    Job,
    JobError,
    JobResult,
    JobUpdateMetadata,
    ParametersModel,
)
from ._results import ResultStore
from ._storage import JobStore

__all__ = ["JobService"]


class JobService:
    """Dispatch and track UWS jobs.

    The goal of this layer is to encapsulate the machinery of a service that
    dispatches jobs using arq, without making assumptions about what the jobs
    do or what outputs they may return. Workers mostly do not use this layer
    and instead talk directly to the `~safir.uws._storage.WorkerJobStore`.

    Parameters
    ----------
    config
        UWS configuration.
    arq_queue
        arq queue to which to dispatch jobs.
    storage
        Underlying storage for job metadata and result tracking.
    logger
        Logger to use.
    """

    def __init__(
        self,
        *,
        config: UWSConfig,
        arq_queue: ArqQueue,
        storage: JobStore,
        logger: BoundLogger,
    ) -> None:
        self._config = config
        self._arq = arq_queue
        self._storage = storage
        self._logger = logger

    async def abort(self, token: str, job_id: str) -> None:
        """Abort a queued or running job.

        If the job is already in a completed state, this operation does
        nothing.

        Parameters
        ----------
        token
            Delegated token for user.
        job_id
            Identifier of the job.

        Raises
        ------
        UnknownJobError
            Raised if the job was not found.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        job = await self._storage.get(token, job_id)
        logger = self._build_logger_for_job(job)
        if job.phase not in ACTIVE_PHASES:
            logger.info(f"Cannot stop job in phase {job.phase.value}")
            return
        if job.message_id:
            timeout = JOB_STOP_TIMEOUT.total_seconds()
            logger.info("Aborting queued job", arq_job_id=job.message_id)
            await self._arq.abort_job(job.message_id, timeout=timeout)
        await self._storage.mark_aborted(token, job_id)
        logger.info("Aborted job")

    async def availability(self) -> Availability:
        """Check the availability of underlying services.

        Currently, this does nothing. Eventually, it may do a health check of
        Wobbly.
        """
        return Availability(available=True)

    async def create(
        self,
        token: str,
        parameters: ParametersModel,
        *,
        run_id: str | None = None,
    ) -> Job:
        """Create a pending job.

        This does not start execution of the job. That must be done separately
        with `start`.

        Parameters
        ----------
        token
            Delegated token for user.
        parameters
            The input parameters to the job.
        run_id
            A client-supplied opaque identifier to record with the job.

        Returns
        -------
        Job
            Information about the newly-created job.

        Raises
        ------
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        job = await self._storage.create(
            token,
            run_id=run_id,
            parameters=parameters,
            execution_duration=self._config.execution_duration,
            lifetime=self._config.lifetime,
        )
        logger = self._build_logger_for_job(job)
        logger.info("Created job")
        return job

    async def delete(self, token: str, job_id: str) -> None:
        """Delete a job.

        If the job is in an active phase, cancel it before deleting it.

        Parameters
        ----------
        token
            Delegated token for user.
        job_id
            Identifier of job.

        Raises
        ------
        UnknownJobError
            Raised if the job was not found.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        job = await self._storage.get(token, job_id)
        logger = self._build_logger_for_job(job)
        if job.phase in ACTIVE_PHASES and job.message_id:
            try:
                await self._arq.abort_job(job.message_id)
            except Exception as e:
                logger.warning("Unable to abort job", error=str(e))
        await self._storage.delete(token, job_id)
        logger.info("Deleted job")

    async def get(
        self,
        token: str,
        job_id: str,
        *,
        wait_seconds: int | None = None,
        wait_phase: ExecutionPhase | None = None,
        wait_for_completion: bool = False,
    ) -> Job:
        """Retrieve a job.

        This also supports long-polling, to implement UWS 1.1 blocking
        behavior, and waiting for completion, to use as a building block when
        constructing a sync API.

        Parameters
        ----------
        token
            Delegated token for user.
        job_id
            Identifier of the job.
        wait_seconds
            If given, wait up to this many seconds for the status to change
            before returning. -1 indicates waiting the maximum length of
            time. This is done by polling the database with exponential
            backoff. This will only be honored if the phase is ``PENDING``,
            ``QUEUED``, or ``EXECUTING``.
        wait_phase
            If ``wait`` was given, the starting phase for waiting. Returns
            immediately if the initial phase doesn't match this one.
        wait_for_completion
            If set to true, wait until the job completes (has a phase other
            than ``QUEUED`` or ``EXECUTING``). Only one of this or
            ``wait_phase`` should be given. Ignored if ``wait`` was not given.

        Returns
        -------
        Job
            Corresponding job.

        Raises
        ------
        UnknownJobError
            Raised if the job was not found.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.

        Notes
        -----
        ``wait_seconds`` and related parameters are relatively inefficient
        since they poll the database using exponential backoff (starting at a
        0.1s delay and increasing by 1.5x). This may need to be reconsidered
        if it becomes a performance bottleneck.
        """
        job = await self._storage.get(token, job_id)

        # If waiting for a status change was requested and is meaningful, do
        # so, capping the wait time at the configured maximum timeout.
        if wait_seconds and job.phase in ACTIVE_PHASES:
            if wait_seconds < 0:
                wait = self._config.wait_timeout
            else:
                wait = timedelta(seconds=wait_seconds)
                wait = min(wait, self._config.wait_timeout)
            if wait_for_completion:
                until_not = ACTIVE_PHASES
            else:
                until_not = {wait_phase} if wait_phase else {job.phase}
            job = await self._wait_for_job(token, job, until_not, timeout=wait)

        return job

    async def get_error(
        self, token: str, job_id: str
    ) -> list[JobError] | None:
        """Get the errors for a job, if any.

        Parameters
        ----------
        token
            Delegated token for user.
        job_id
            Identifier of the job.

        Returns
        -------
        list of JobError or None
            Error information for the job, or `None` if the job didn't fail.

        Raises
        ------
        UnknownJobError
            Raised if the job was not found.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        job = await self._storage.get(token, job_id)
        return job.errors

    async def get_summary(
        self,
        token: str,
        job_id: str,
        *,
        signer: ResultStore | None = None,
        wait_seconds: int | None = None,
        wait_phase: ExecutionPhase | None = None,
    ) -> JobSummary:
        """Retrieve a job's XML model.

        This also supports long-polling, to implement UWS 1.1 blocking
        behavior, and signing of results.

        Parameters
        ----------
        token
            Delegated token for user.
        job_id
            Identifier of the job.
        signer
            If provided, generate signed URLs for the results by using the
            given signing object. If no object is supplied, no URLs will be
            included in the result list.
        wait_seconds
            If given, wait up to this many seconds for the status to change
            before returning. -1 indicates waiting the maximum length of
            time. This is done by polling the database with exponential
            backoff. This will only be honored if the phase is ``PENDING``,
            ``QUEUED``, or ``EXECUTING``.
        wait_phase
            If ``wait`` was given, the starting phase for waiting. Returns
            immediately if the initial phase doesn't match this one.

        Returns
        -------
        JobSummary
            Corresponding job.

        Raises
        ------
        UnknownJobError
            Raised if the job was not found.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.

        Notes
        -----
        ``wait_seconds`` and related parameters are relatively inefficient
        since they poll the database using exponential backoff (starting at a
        0.1s delay and increasing by 1.5x). This may need to be reconsidered
        if it becomes a performance bottleneck.
        """
        job = await self.get(
            token, job_id, wait_seconds=wait_seconds, wait_phase=wait_phase
        )
        if signer:
            job.results = [signer.sign_url(r) for r in job.results]
        return job.to_xml_model(self._config.job_summary_type)

    async def list_jobs(
        self,
        token: str,
        base_url: str,
        *,
        phases: list[ExecutionPhase] | None = None,
        after: datetime | None = None,
        count: int | None = None,
    ) -> Jobs:
        """List the jobs for a particular user.

        Parameters
        ----------
        token
            Delegated token for user.
        base_url
            Base URL used to form URLs to the specific jobs.
        phases
            Limit the result to jobs in this list of possible execution
            phases.
        after
            Limit the result to jobs created after the given datetime.
        count
            Limit the results to the most recent count jobs.

        Returns
        -------
        Jobs
            Collection of short job descriptions.

        Raises
        ------
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        jobs = await self._storage.list_jobs(
            token, phases=phases, after=after, count=count
        )
        return Jobs(jobref=[j.to_job_description(base_url) for j in jobs])

    async def run_sync(
        self,
        token: str,
        user: str,
        parameters: ParametersModel,
        *,
        runid: str | None,
    ) -> JobResult:
        """Create a job for a sync request and return the first result.

        Parameters
        ----------
        token
            Delegated token for user.
        user
            User on behalf of whom this operation is performed.
        params
            Job parameters.
        user
            Username of user running the job.
        runid
            User-supplied RunID, if any.

        Returns
        -------
        JobResult
            First result of the successfully-executed job.

        Raises
        ------
        SyncJobFailedError
            Raised if the job failed.
        SyncJobNoResultsError
            Raised if the job returned no results.
        SyncJobTimeoutError
            Raised if the job execution timed out.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        job = await self.create(token, parameters, run_id=runid)
        logger = self._build_logger_for_job(job)

        # Start the job and wait for it to complete.
        metadata = await self.start(token, user, job.id)
        logger = logger.bind(arq_job_id=metadata.id)
        job = await self.get(
            token,
            job.id,
            wait_seconds=int(self._config.sync_timeout.total_seconds()),
            wait_for_completion=True,
        )

        # Check for error states.
        if job.phase not in (ExecutionPhase.COMPLETED, ExecutionPhase.ERROR):
            logger.warning("Job timed out", timeout=self._config.sync_timeout)
            raise SyncJobTimeoutError(self._config.sync_timeout)
        if job.errors:
            # Only one error is supported for right now.
            error = job.errors[0]
            logger.warning("Job failed", error=error.model_dump(mode="json"))
            raise SyncJobFailedError(error)
        if not job.results:
            logger.warning("Job returned no results")
            raise SyncJobNoResultsError

        # Return the first result.
        return job.results[0]

    async def start(self, token: str, user: str, job_id: str) -> JobMetadata:
        """Start execution of a job.

        Parameters
        ----------
        token
            Gafaelfawr token used to authenticate to services used by the
            backend on the user's behalf.
        user
            User on behalf of whom this operation is performed.
        job_id
            Identifier of the job to start.

        Returns
        -------
        JobMetadata
            arq job metadata.

        Raises
        ------
        UnknownJobError
            Raised if the job was not found.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        job = await self._storage.get(token, job_id)
        if job.phase not in (ExecutionPhase.PENDING, ExecutionPhase.HELD):
            raise InvalidPhaseError(f"Cannot start job in phase {job.phase}")
        logger = self._build_logger_for_job(job)
        info = WorkerJobInfo(
            job_id=job.id,
            user=user,
            token=token,
            timeout=job.execution_duration or self._config.lifetime,
            run_id=job.run_id,
        )
        params = job.parameters.to_worker_parameters().model_dump(mode="json")
        metadata = await self._arq.enqueue(self._config.worker, params, info)
        await self._storage.mark_queued(token, job_id, metadata)
        logger.info("Started job", arq_job_id=metadata.id)
        return metadata

    async def update_destruction(
        self, token: str, job_id: str, destruction: datetime
    ) -> datetime | None:
        """Update the destruction time of a job.

        Parameters
        ----------
        token
            Delegated token for user.
        job_id
            Identifier of the job to update.
        destruction
            The new job destruction time. This may be modified by the service
            callback if it so desires.

        Returns
        -------
        datetime.datetime or None
            The new destruction time of the job (possibly modified by the
            callback), or `None` if the destruction time of the job was
            not changed.

        Raises
        ------
        UnknownJobError
            Raised if the job was not found.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        job = await self._storage.get(token, job_id)
        logger = self._build_logger_for_job(job)

        # Validate the new value.
        if validator := self._config.validate_destruction:
            destruction = validator(destruction, job)
        elif destruction > job.destruction_time:
            destruction = job.destruction_time

        # Update the destruction time if needed.
        if destruction == job.destruction_time:
            return None
        metadata = JobUpdateMetadata(
            destruction_time=destruction,
            execution_duration=job.execution_duration,
        )
        await self._storage.update_metadata(token, job_id, metadata)
        logger.info(
            "Changed job destruction time",
            destruction=isodatetime(destruction),
        )
        return destruction

    async def update_execution_duration(
        self, token: str, job_id: str, duration: timedelta | None
    ) -> timedelta | None:
        """Update the execution duration time of a job.

        Parameters
        ----------
        token
            Delegated token for user.
        job_id
            Identifier of the job to update.
        duration
            The new job execution duration. This may be modified by the service
            callback if it so desires.

        Returns
        -------
        timedelta or None
            The new execution duration of the job (possibly modified by the
            callback), or `None` if the execution duration of the job was
            not changed.

        Raises
        ------
        UnknownJobError
            Raised if the job was not found.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        job = await self._storage.get(token, job_id)
        logger = self._build_logger_for_job(job)

        # Validate the new value.
        if validator := self._config.validate_execution_duration:
            duration = validator(duration, job)
        if duration:
            duration = min(duration, self._config.execution_duration)

        # Update the duration in the job.
        if duration == job.execution_duration:
            return None
        update = JobUpdateMetadata(
            destruction_time=job.destruction_time, execution_duration=duration
        )
        await self._storage.update_metadata(token, job_id, update)
        if duration:
            duration_str = f"{duration.total_seconds()}s"
        else:
            duration_str = "unlimited"
        logger.info("Changed job execution duration", duration=duration_str)
        return duration

    def _build_logger_for_job(self, job: Job) -> BoundLogger:
        """Construct a logger with bound information for a job.

        Parameters
        ----------
        job
            Job for which to report messages.

        Returns
        -------
        BoundLogger
            Logger with more bound metadata.
        """
        logger = self._logger.bind(user=job.owner, job_id=job.id)
        if job.run_id:
            logger = logger.bind(run_id=job.run_id)
        if job.parameters:
            parameters = job.parameters.model_dump(mode="json")
            logger = logger.bind(parameters=parameters)
        return logger

    async def _wait_for_job(
        self,
        token: str,
        job: Job,
        until_not: set[ExecutionPhase],
        *,
        timeout: timedelta,
    ) -> Job:
        """Wait for the completion of a job.

        Parameters
        ----------
        token
            Delegated token for user.
        job
            Job to wait for.
        until_not
            Wait until the job is no longer in one of this set of phases.
        timeout
            How long to wait.

        Returns
        -------
        Job
            The new state of the job.
        """
        # I don't know of a way to set a watch on the database, so use
        # polling. Poll the database with exponential delay starting with 0.1
        # seconds and increasing by 1.5x each time until we reach the maximum
        # duration.
        delay = 0.1
        max_delay = 5
        with suppress(TimeoutError):
            async with asyncio.timeout(timeout.total_seconds()):
                while job.phase in until_not:
                    await asyncio.sleep(delay)
                    job = await self._storage.get(token, job.id)
                    delay = min(delay * 1.5, max_delay)

        # If we timed out, we may have done so in the middle of a delay. Try
        # one last request.
        return await self._storage.get(token, job.id)



"""Update the request based on ``X-Forwarded-For`` headers."""

from __future__ import annotations

from copy import copy
from ipaddress import _BaseAddress, _BaseNetwork, ip_address

from starlette.datastructures import Headers
from starlette.types import ASGIApp, Receive, Scope, Send

__all__ = ["XForwardedMiddleware"]


class XForwardedMiddleware:
    """ASGI middleware to update the request based on ``X-Forwarded-For``.

    The remote IP address will be replaced with the right-most IP address in
    ``X-Forwarded-For`` that is not contained within one of the trusted
    proxy networks.

    If ``X-Forwarded-For`` is found and ``X-Forwarded-Proto`` is also present,
    the corresponding entry of ``X-Forwarded-Proto`` is used to replace the
    scheme in the request scope. If ``X-Forwarded-Proto`` only has one entry
    (ingress-nginx has this behavior), that one entry will become the new
    scheme in the request scope.

    The contents of ``X-Forwarded-Host`` will be stored as ``forwarded_host``
    in the request state if it and ``X-Forwarded-For`` are present. Normally
    this is not needed since NGINX will pass the original ``Host`` header
    without modification.

    Parameters
    ----------
    proxies
        The networks of the trusted proxies. If not specified, defaults to the
        empty list, which means only the immediately upstream proxy will be
        trusted.
    """

    def __init__(
        self, app: ASGIApp, *, proxies: list[_BaseNetwork] | None = None
    ) -> None:
        self._app = app
        self._proxies = proxies if proxies else []

    async def __call__(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None:
        if scope["type"] != "http":
            await self._app(scope, receive, send)
            return
        scope = copy(scope)
        scope.setdefault("state", {})
        headers = Headers(scope=scope)
        forwarded_for = list(reversed(self._get_forwarded_for(headers)))
        if not forwarded_for:
            scope["state"]["forwarded_host"] = None
            await self._app(scope, receive, send)
            return

        client = None
        for n, ip in enumerate(forwarded_for):
            if any(ip in network for network in self._proxies):
                continue
            client = str(ip)
            index = n
            break

        # If all the IP addresses are from trusted networks, take the
        # left-most.
        if not client:
            client = str(forwarded_for[-1])
            index = -1

        # Update the request's understanding of the client IP.
        if scope.get("client"):
            scope["client"] = (client, scope["client"][1])
        else:
            scope["client"] = (client, None)

        # Ideally this should take the scheme corresponding to the entry in
        # X-Forwarded-For that was chosen, but some proxies (the Kubernetes
        # NGINX ingress, for example) only retain one element in
        # X-Forwarded-Proto. In that case, use what we have.
        proto = list(reversed(self._get_forwarded_proto(headers)))
        if proto:
            if index >= len(proto):
                index = -1
            scope["scheme"] = proto[index]

        # Record what appears to be the client host for logging purposes.
        scope["state"]["forwarded_host"] = self._get_forwarded_host(headers)

        # Perform the rest of the request processing.
        await self._app(scope, receive, send)
        return

    def _get_forwarded_for(self, headers: Headers) -> list[_BaseAddress]:
        """Retrieve the ``X-Forwarded-For`` entries from the request.

        Parameters
        ----------
        scope
            Request headers.

        Returns
        -------
        list of ipaddress._BaseAddress
            The list of addresses found in the header. If there are multiple
            ``X-Forwarded-For`` headers, we don't know which one is correct,
            so act as if there are no headers.
        """
        forwarded_for_str = headers.getlist("X-Forwarded-For")
        if not forwarded_for_str or len(forwarded_for_str) > 1:
            return []
        return [
            ip_address(addr)
            for addr in (a.strip() for a in forwarded_for_str[0].split(","))
            if addr
        ]

    def _get_forwarded_host(self, headers: Headers) -> str | None:
        """Retrieve the ``X-Forwarded-Host`` header.

        Parameters
        ----------
        headers
            Request headers.

        Returns
        -------
        str
            The value of the ``X-Forwarded-Host`` header, if present and if
            there is only one header. If there are multiple
            ``X-Forwarded-Host`` headers, we don't know which one is correct,
            so act as if there are no headers.
        """
        forwarded_host = headers.getlist("X-Forwarded-Host")
        if not forwarded_host or len(forwarded_host) > 1:
            return None
        return forwarded_host[0].strip()

    def _get_forwarded_proto(self, headers: Headers) -> list[str]:
        """Retrieve the ``X-Forwarded-Proto`` entries from the request.

        Parameters
        ----------
        headers
            Request headers.

        Returns
        -------
        list of str
            The list of schemes found in the header. If there are multiple
            ``X-Forwarded-Proto`` headers, we don't know which one is correct,
            so act as if there are no headers.
        """
        forwarded_proto_str = headers.getlist("X-Forwarded-Proto")
        if not forwarded_proto_str or len(forwarded_proto_str) > 1:
            return []
        return [p.strip() for p in forwarded_proto_str[0].split(",")]



"""Help command implementation for Click."""

from __future__ import annotations

import click

__all__ = ["display_help"]


def display_help(
    main: click.Group,
    ctx: click.Context,
    topic: str | None = None,
    subtopic: str | None = None,
) -> None:
    """Show help for a Click command.

    This function implements a help command for command-line interfaces using
    Click_. It supports commands and subcommands, displaying the same
    information as would be shown by ``--help``. The resulting help text is
    displayed with `click.echo`.

    Parameters
    ----------
    main
        Top-level Click command group.
    ctx
        Click context.
    topic
        Command for which to get help.
    subtopic
        Subcommand of that command for which to get help.

    Raises
    ------
    click.UsageError
        Raised if the given help topic or subtopic does not correspond to a
        registered Click command.
    RuntimeError
        Raised if the Click context has no parent, meaning that this was
        called with an invalid context or from the top-level Click command.

    Returns
    -------
    `None`

    Examples
    --------
    This function should normally be called from a top-level ``help`` command.
    Assuming that the top-level Click group is named ``main``, here is the
    typical usage:

    .. code-block:: python

       @main.command()
       @click.argument("topic", default=None, required=False, nargs=1)
       @click.argument("subtopic", default=None, required=False, nargs=1)
       @click.pass_context
       def help(
           ctx: click.Context, topic: str | None, subtopic: str | None
       ) -> None:
           display_help(main, ctx, topic, subtopic)
    """
    if not topic:
        if not ctx.parent:
            raise RuntimeError("help called without topic or parent")
        click.echo(ctx.parent.get_help())
        return
    if topic not in main.commands:
        raise click.UsageError(f"Unknown help topic {topic}", ctx)
    if not subtopic:
        ctx.info_name = topic
        click.echo(main.commands[topic].get_help(ctx))
        return

    # Subtopic handling. This requires some care with typing, since the
    # commands attribute (although present) is not documented, and the
    # get_command method is only available on Groups.
    group = main.commands[topic]
    if isinstance(group, click.Group):
        command = group.get_command(ctx, subtopic)
        if command:
            ctx.info_name = f"{topic} {subtopic}"
            click.echo(command.get_help(ctx))
            return

    # Fall through to the error case of no subcommand found.
    msg = f"Unknown help topic {topic} {subtopic}"
    raise click.UsageError(msg, ctx)



"""Test Kafka exception functionality."""

from collections.abc import AsyncGenerator
from unittest.mock import ANY

import pytest
import pytest_asyncio
from faststream.kafka import KafkaBroker
from faststream.kafka.testing import TestKafkaBroker
from structlog.stdlib import get_logger

from safir.kafka import FastStreamErrorHandler
from safir.testing.sentry import Captured
from safir.testing.slack import MockSlackWebhook


@pytest_asyncio.fixture
async def error_handling_faststream_broker(
    faststream_error_handler: FastStreamErrorHandler,
) -> AsyncGenerator[KafkaBroker]:
    """Yield A FastStream Kafka broker with error reporting middleware."""
    broker = KafkaBroker(
        middlewares=[faststream_error_handler.make_middleware()]
    )

    @broker.subscriber("test-with-errors")
    async def handler_with_error(foo: str) -> None:
        _ = 1 / 0

    @broker.subscriber("test-without-errors")
    async def handler_without_error(foo: str) -> None:
        return

    async with TestKafkaBroker(broker) as br:
        yield br


@pytest.mark.asyncio
async def test_faststream_error_handler_doesnt_error(
    error_handling_faststream_broker: KafkaBroker,
) -> None:
    broker = error_handling_faststream_broker
    await broker.publish("whatever", topic="test-without-errors")


@pytest.mark.asyncio
async def test_faststream_error_handler_reraises(
    error_handling_faststream_broker: KafkaBroker,
) -> None:
    broker = error_handling_faststream_broker
    with pytest.raises(ZeroDivisionError):
        await broker.publish("whatever", topic="test-with-errors")


@pytest.mark.asyncio
async def test_faststream_error_handler_reports_slack(
    error_handling_faststream_broker: KafkaBroker,
    faststream_error_handler: FastStreamErrorHandler,
    mock_slack: MockSlackWebhook,
) -> None:
    faststream_error_handler.initialize_slack(
        hook_url=mock_slack.url,
        application="my-app",
        logger=get_logger(),
    )
    broker = error_handling_faststream_broker
    with pytest.raises(ZeroDivisionError):
        await broker.publish("whatever", topic="test-with-errors")

    expected = [
        {
            "blocks": [
                {
                    "text": {
                        "text": "Error in my-app",
                        "type": "mrkdwn",
                        "verbatim": True,
                    },
                    "type": "section",
                },
                {
                    "fields": [
                        {
                            "text": "*Exception type*\nZeroDivisionError",
                            "type": "mrkdwn",
                            "verbatim": True,
                        },
                        {
                            "text": ANY,
                            "type": "mrkdwn",
                            "verbatim": True,
                        },
                    ],
                    "type": "section",
                },
                {
                    "text": {
                        "text": "*Exception*\n"
                        "```\n"
                        "ZeroDivisionError: division by zero\n"
                        "```",
                        "type": "mrkdwn",
                        "verbatim": True,
                    },
                    "type": "section",
                },
                {
                    "type": "divider",
                },
            ],
        },
    ]

    assert mock_slack.messages == expected


@pytest.mark.asyncio
async def test_faststream_error_handler_reports_sentry(
    error_handling_faststream_broker: KafkaBroker,
    sentry_combo_items: Captured,
) -> None:
    broker = error_handling_faststream_broker
    with pytest.raises(ZeroDivisionError):
        await broker.publish("whatever", topic="test-with-errors")
    (error,) = sentry_combo_items.errors
    (exception,) = error["exception"]["values"]
    assert exception["type"] == "ZeroDivisionError"



"""Define arq workers for an application using UWS."""

from __future__ import annotations

import asyncio
import contextlib
import uuid
from datetime import UTC, datetime
from typing import Any, ParamSpec

from httpx import AsyncClient
from structlog.stdlib import BoundLogger

from safir.sentry._helpers import report_exception

try:
    from safir.arq import (
        ArqMode,
        ArqQueue,
        JobNotFound,
        JobResult,
        JobResultUnavailable,
        MockArqQueue,
        RedisArqQueue,
    )
    from safir.arq.uws import WorkerError, WorkerTransientError
except ImportError as e:
    raise ImportError(
        "The safir.uws module requires the uws extra. "
        "Install it with `pip install safir[uws]`."
    ) from e
from safir.datetime import format_datetime_for_logging
from safir.dependencies.http_client import http_client_dependency
from safir.slack.blockkit import SlackException
from safir.slack.webhook import SlackWebhookClient

from ._config import UWSConfig
from ._constants import JOB_RESULT_TIMEOUT, WOBBLY_REQUEST_TIMEOUT
from ._exceptions import TaskError, UnknownJobError
from ._models import Job
from ._storage import JobStore

P = ParamSpec("P")

__all__ = [
    "close_uws_worker_context",
    "create_uws_worker_context",
    "uws_job_completed",
    "uws_job_started",
]


async def create_uws_worker_context(
    config: UWSConfig, logger: BoundLogger
) -> dict[str, Any]:
    """Construct the arq context for UWS workers.

    The return value is a dictionary that should be merged with the ``ctx``
    dictionary that is passed to each worker.

    Parameters
    ----------
    config
        UWS configuration.
    logger
        Logger for the worker to use.

    Returns
    -------
    dict
        Keys to add to the ``ctx`` dictionary.
    """
    logger = logger.bind(worker_instance=uuid.uuid4().hex)

    # The queue from which to retrieve results is the main work queue,
    # which uses the default arq queue name. Note that this is not the
    # separate UWS queue this worker is running against.
    if config.arq_mode == ArqMode.production:
        settings = config.arq_redis_settings
        arq: ArqQueue = await RedisArqQueue.initialize(settings)
    else:
        arq = MockArqQueue()

    http_client = AsyncClient(timeout=WOBBLY_REQUEST_TIMEOUT)
    storage = JobStore(config, http_client)
    slack = None
    if config.slack_webhook:
        slack = SlackWebhookClient(
            config.slack_webhook.get_secret_value(),
            "vo-cutouts-db-worker",
            logger,
        )

    logger.info("Worker startup complete")
    return {
        "arq": arq,
        "http_client": http_client,
        "logger": logger,
        "slack": slack,
        "storage": storage,
    }


async def close_uws_worker_context(ctx: dict[Any, Any]) -> None:
    """Close the context used by the UWS workers.

    Performs any necessary cleanup of persistent objects stored in the ``ctx``
    argument passed to each worker.

    Parameters
    ----------
    ctx
        Worker context.
    """
    logger: BoundLogger = ctx["logger"]
    http_client: AsyncClient = ctx["http_client"]

    await http_client.aclose()

    # Possibly initialized by the Slack webhook client.
    await http_client_dependency.aclose()

    logger.info("Worker shutdown complete")


async def uws_job_started(
    ctx: dict[Any, Any], token: str, job_id: str, start_time: datetime
) -> None:
    """Mark a UWS job as executing.

    Parameters
    ----------
    ctx
        arq context.
    token
        Token for the user executing the job.
    job_id
        UWS job identifier.
    start_time
        When the job was started.
    """
    logger: BoundLogger = ctx["logger"].bind(task="job_started", job_id=job_id)
    slack: SlackWebhookClient | None = ctx["slack"]
    storage: JobStore = ctx["storage"]

    try:
        await storage.mark_executing(token, job_id, start_time)
        logger.info(
            "Marked job as started",
            start_time=format_datetime_for_logging(start_time),
        )
    except UnknownJobError:
        logger.warning("Job not found to mark as started", job_id=job_id)
    except Exception as e:
        # We don't use safir.slack.report_exception here because we want the
        # exception to propagate to the Sentry uncaught exception
        # handler so that we get the full instrumentation from any configured
        # Sentry integrations.
        if slack:
            await slack.post_exception(e)
        raise


async def _annotate_worker_error(
    exc: Exception, job: Job, slack: SlackWebhookClient | None = None
) -> Exception:
    """Convert and possibly report a backend worker error.

    Convert the backend worker error to a task error and annotate it with task
    information. Report the error to Slack if Slack is configured and the
    error is not ignored for Slack reporting purposes.

    Parameters
    ----------
    exc
        Worker exception.
    job
        Associated UWS job.
    slack
        Class for reporting errors to Slack, if Slack error reporting is
        enabled.

    Returns
    -------
    TaskError
        Exception converted to a `~safir.uws._exceptions.TaskError`.
    """
    match exc:
        case WorkerError():
            error = TaskError.from_worker_error(exc)
            error.job_id = job.id
            error.started_at = job.creation_time
            error.user = job.owner
            if not error.slack_ignore:
                await report_exception(error, slack_client=slack)
            return error
        case SlackException():
            exc.user = job.owner

    await report_exception(exc, slack_client=slack)
    return exc


async def _get_job_result(arq: ArqQueue, arq_job_id: str) -> JobResult:
    """Get the result of the job, which may require waiting for it."""
    now = datetime.now(tz=UTC)
    end = now + JOB_RESULT_TIMEOUT
    while now < end:
        await asyncio.sleep(0.5)
        with contextlib.suppress(JobResultUnavailable):
            return await arq.get_job_result(arq_job_id)

    # If we fell off the end of the retry loop, try one more time.
    return await arq.get_job_result(arq_job_id)


async def uws_job_completed(
    ctx: dict[Any, Any], token: str, job_id: str
) -> None:
    """Mark a UWS job as completed.

    Recover the exception if the job failed and record that as the job error.
    Because we can't use the arq ``after_job_end`` callback, the job results
    may not be available yet when we're called, which requires polling.

    Parameters
    ----------
    ctx
        arq context.
    token
        Token for the user executing the job.
    job_id
        UWS job identifier.
    """
    arq: ArqQueue = ctx["arq"]
    logger: BoundLogger = ctx["logger"].bind(
        task="job_completed", job_id=job_id
    )
    slack: SlackWebhookClient | None = ctx["slack"]
    storage: JobStore = ctx["storage"]

    try:
        job = await storage.get(token, job_id)
        arq_job_id = job.message_id
        if not arq_job_id:
            msg = "Job has no associated arq job ID, cannot mark completed"
            logger.error(msg)
            return
        logger = logger.bind(arq_job_id=arq_job_id)

        # Get the job results.
        try:
            result = await _get_job_result(arq, arq_job_id)
        except (JobNotFound, JobResultUnavailable) as e:
            logger.exception("Cannot retrieve job result")
            exc = WorkerTransientError(
                "Cannot retrieve job result from job queue",
                f"{type(e).__name__}: {e!s}",
                add_traceback=True,
            )
            exc.__cause__ = e
            error = await _annotate_worker_error(exc, job, slack)
            await storage.mark_failed(token, job_id, error)
            return

        # If the job failed and Slack reporting is enabled, annotate the job
        # with some more details and report it to Slack.
        if isinstance(result.result, Exception):
            error = await _annotate_worker_error(result.result, job, slack)
            result.result = error

        # Mark the job as completed.
        await storage.mark_completed(token, job_id, result)
        logger.info("Marked job as completed")
    except UnknownJobError:
        logger.warning("Job not found to mark as completed")
    except Exception as e:
        # We don't use safir.slack.report_exception here because we want the
        # exception to propagate to the Sentry uncaught exception
        # handler so that we get the full instrumentation from any configured
        # Sentry integrations.
        if slack:
            await slack.post_exception(e)
        raise



"""Store Pydantic models in Redis with optional encryption."""

from collections.abc import AsyncGenerator
from typing import override

try:
    import redis.asyncio as redis
except ImportError as e:
    raise ImportError(
        "The safir.redis module requires the redis extra. "
        "Install it with `pip install safir[redis]`."
    ) from e
from cryptography.fernet import Fernet
from pydantic import BaseModel, SecretStr

from safir.slack.blockkit import (
    SentryEventInfo,
    SlackCodeBlock,
    SlackException,
    SlackMessage,
    SlackTextField,
)

__all__ = [
    "DeserializeError",
    "EncryptedPydanticRedisStorage",
    "PydanticRedisStorage",
]


class DeserializeError(SlackException):
    """Raised when a stored Pydantic object in Redis cannot be decoded (and
    possibly decrypted) or deserialized.

    Parameters
    ----------
    msg
        A description of the error.
    key
        The key of the object that could not be deserialized.
    error
        The exception's message from the deserialization attempt.
    """

    def __init__(self, msg: str, key: str, error: str) -> None:
        super().__init__(msg)
        self.key = key
        self.error = error

    @override
    def to_slack(self) -> SlackMessage:
        message = super().to_slack()
        message.fields.append(SlackTextField(heading="Key", text=self.key))
        message.blocks.append(SlackCodeBlock(heading="Error", code=self.error))
        return message

    @override
    def to_sentry(self) -> SentryEventInfo:
        info = super().to_sentry()
        info.tags["key"] = self.key
        return info


class PydanticRedisStorage[S: BaseModel]:
    """JSON-serialized encrypted storage in Redis.

    Parameters
    ----------
    datatype
        The class of object being stored (a Pydantic model).
    redis
        A Redis client configured to talk to the backend store.
    key_prefix
        A prefix to prepend to all Redis keys. Setting a prefix ensures that
        all objects stored through this class share the same redis key
        prefix. This is useful if multiple storage classes share the same
        Redis database. The keys that you use to store objects (e.g. the
        `store` method) and the key you get back from the `scan` method do
        not include this prefix.
    """

    def __init__(
        self, *, datatype: type[S], redis: redis.Redis, key_prefix: str = ""
    ) -> None:
        self._datatype = datatype
        self._redis = redis
        self._key_prefix = key_prefix

    async def delete(self, key: str) -> bool:
        """Delete a stored object.

        Parameters
        ----------
        key
            The key to delete.

        Returns
        -------
        bool
            `True` if the key was found and deleted, `False` otherwise.
        """
        count = await self._redis.delete(self._prefix_key(key))
        return count > 0

    async def delete_all(self, pattern: str) -> None:
        """Delete all stored objects.

        Parameters
        ----------
        pattern
            Glob pattern matching the keys to purge, such as ``oidc:*``.
        """
        async for key in self._redis.scan_iter(self._prefix_key(pattern)):
            await self._redis.delete(key)

    async def get(self, key: str) -> S | None:
        """Retrieve a stored object.

        Parameters
        ----------
        key
            The key for the object.

        Returns
        -------
        Any or None
            The deserialized object or `None` if no such object could be
            found.

        Raises
        ------
        DeserializeError
            Raised if the stored object could not be decrypted or
            deserialized.
        """
        full_key = self._prefix_key(key)
        data = await self._redis.get(full_key)
        if not data:
            return None

        try:
            return self._deserialize(data)
        except Exception as e:
            error = f"{type(e).__name__}: {e!s}"
            msg = f"Cannot deserialize data for key {full_key}"
            raise DeserializeError(msg, key=full_key, error=error) from e

    async def scan(self, pattern: str) -> AsyncGenerator[str]:
        """Scan Redis for a given key pattern, returning each key.

        Parameters
        ----------
        pattern
            Key pattern to scan for.

        Yields
        ------
        str
            Each key matching that pattern.
        """
        async for key in self._redis.scan_iter(
            match=self._prefix_key(pattern)
        ):
            yield key.decode().removeprefix(self._key_prefix)

    async def store(
        self,
        key: str,
        obj: S,
        lifetime: int | None = None,
        *,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> None:
        """Store an object.

        Parameters
        ----------
        key
            The key for the object.
        obj
            The object to store.
        lifetime
            The object lifetime in seconds.  The object should expire from the
            data store after that many seconds after the current time.  Pass
            `None` if the object should not expire.
        exclude_unset
            If `True`, exclude fields that have not been explicitly set.
        exclude_defaults
            If `True`, exclude fields set to their default value.
        exclude_none
            If `True`, exclude fields that have a value of `None`.
        """
        data = self._serialize(
            obj,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        await self._redis.set(self._prefix_key(key), data, ex=lifetime)

    def _prefix_key(self, key: str) -> str:
        """Compute the full Redis key, given the key prefix."""
        return f"{self._key_prefix}{key}"

    def _serialize(
        self,
        obj: S,
        *,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> bytes:
        """Serialize a Pydantic object to bytes that can be stored by Redis.

        Parameters
        ----------
        obj
            The Pydantic object to serialize.
        exclude_unset
            If `True`, exclude fields that have not been explicitly set.
        exclude_defaults
            If `True`, exclude fields set to their default value.
        exclude_none
            If `True`, exclude fields that have a value of `None`.

        Returns
        -------
        bytes
            The serialized object.
        """
        return obj.model_dump_json(
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        ).encode()

    def _deserialize(self, data: bytes) -> S:
        """Deserialize bytes into a Pydantic object.

        Parameters
        ----------
        data
            The data to deserialize.

        Returns
        -------
        S
            The deserialized Pydantic object.
        """
        return self._datatype.model_validate_json(data.decode())


class EncryptedPydanticRedisStorage[S: BaseModel](PydanticRedisStorage[S]):
    """A Pydantic-based Redis store that encrypts data.

    Parameters
    ----------
    datatype
        The class of object being stored (a Pydantic model).
    redis
        A Redis client configured to talk to the backend store.
    encryption_key
        Encryption key. Must be a `~cryptography.fernet.Fernet` key. Generate
        a key with ``Fernet.generate_key().decode()``.
    key_prefix
        A prefix to prepend to all Redis keys. Setting a prefix ensures that
        all objects stored through this class share the same redis key
        prefix. This is useful if multiple storage classes share the same
        Redis database. The keys that you use to store objects (e.g. the
        `store` method) and the key you get back from the `scan` method do
        not include this prefix.
    """

    def __init__(
        self,
        *,
        datatype: type[S],
        redis: redis.Redis,
        encryption_key: str | SecretStr,
        key_prefix: str = "",
    ) -> None:
        super().__init__(datatype=datatype, redis=redis, key_prefix=key_prefix)
        if isinstance(encryption_key, SecretStr):
            encryption_key = encryption_key.get_secret_value()
        self._fernet = Fernet(encryption_key.encode())

    @override
    def _serialize(
        self,
        obj: S,
        *,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> bytes:
        data = super()._serialize(
            obj,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        return self._fernet.encrypt(data)

    @override
    def _deserialize(self, data: bytes) -> S:
        data = self._fernet.decrypt(data)
        return self._datatype.model_validate_json(data.decode())



"""Mock Kubernetes API and utility functions for testing."""

from ._mock import MockKubernetesApi, patch_kubernetes
from ._strip import strip_none

__all__ = [
    "MockKubernetesApi",
    "patch_kubernetes",
    "strip_none",
]



"""Send messages to Slack."""

from __future__ import annotations

from collections.abc import Callable, Coroutine
from typing import Any, ClassVar, override

from fastapi import HTTPException, Request, Response
from fastapi.exceptions import RequestValidationError
from fastapi.routing import APIRoute
from pydantic import SecretStr
from starlette.exceptions import HTTPException as StarletteHTTPException
from structlog.stdlib import BoundLogger

from safir.datetime import current_datetime, format_datetime_for_logging
from safir.dependencies.http_client import http_client_dependency

from .blockkit import (
    SlackCodeBlock,
    SlackException,
    SlackMessage,
    SlackTextField,
)

__all__ = [
    "SlackIgnoredException",
    "SlackRouteErrorHandler",
    "SlackWebhookClient",
]


class SlackIgnoredException(Exception):
    """Parent class for exceptions that should not be reported to Slack.

    This exception has no built-in behavior or meaning except to suppress
    Slack notifications if it is thrown uncaught.  Application exceptions that
    should not result in a Slack alert (because, for example, they're intended
    to be caught by exception handlers) should inherit from this class.
    """


class SlackWebhookClient:
    """Send messages to a Slack webhook.

    Provides a simple Slack client to publish structured messages to a Slack
    channel via an incoming webhook using the Block Kit API. This is a
    write-only client and cannot be used for prompting or more complex apps.

    Parameters
    ----------
    hook_url
        URL of the Slack incoming webhook to use for publishing messages.
    application
        Name of the application reporting an error, used by the error
        reporting methods.
    logger
        Logger to which to report errors sending messages to Slack.
    """

    def __init__(
        self, hook_url: str | SecretStr, application: str, logger: BoundLogger
    ) -> None:
        if isinstance(hook_url, SecretStr):
            self._hook_url = hook_url.get_secret_value()
        else:
            self._hook_url = hook_url
        self._application = application
        self._logger = logger

    async def post(self, message: SlackMessage) -> None:
        """Post a message to Slack.

        Any exceptions encountered while posting the message will be logged
        but not raised. From the perspective of the caller, the message will
        silently disappear.

        Parameters
        ----------
        message
            Message to post.
        """
        self._logger.debug("Sending message to Slack")
        body = message.to_slack()
        try:
            client = await http_client_dependency()
            r = await client.post(self._hook_url, json=body)
            r.raise_for_status()
        except Exception:
            msg = "Posting Slack message failed"
            self._logger.exception(msg, message=body)

    async def post_exception(self, exc: Exception) -> None:
        """Post an alert to Slack about an exception.

        If the exception is a `~safir.slack.blockkit.SlackException`, then the
        Slack message will be constructed by calling it's
        `~safir.slack.blockkit.SlackException.to_slack` method. If the
        exception is not a `~safir.slack.blockkit.SlackException`, then a
        generic Slack message will be constructed.

        This method intentionally does not provide a way to include the
        traceback, since it's hard to ensure that the traceback is entirely
        free of secrets or other information that should not be disclosed on
        Slack. Only the exception message is reported.

        Parameters
        ----------
        exc
            The exception to report.
        """
        if isinstance(exc, SlackException):
            message = exc.to_slack()
            msg = f"Error in {self._application}: {message.message}"
            message.message = msg
        else:
            date = format_datetime_for_logging(current_datetime())
            name = type(exc).__name__
            error = f"{name}: {exc!s}"
            message = SlackMessage(
                message=f"Error in {self._application}",
                fields=[
                    SlackTextField(heading="Exception type", text=name),
                    SlackTextField(heading="Failed at", text=date),
                ],
                blocks=[SlackCodeBlock(heading="Exception", code=error)],
            )
        await self.post(message)


class SlackRouteErrorHandler(APIRoute):
    """Custom `fastapi.routing.APIRoute` that reports exceptions to Slack.

    Dynamically wrap FastAPI route handlers in an exception handler that
    reports uncaught exceptions (other than :exc:`fastapi.HTTPException`,
    :exc:`fastapi.exceptions.RequestValidationError`,
    :exc:`starlette.exceptions.HTTPException`, and exceptions inheriting from
    `SlackIgnoredException`) to Slack.

    This class must be initialized by calling its `initialize` method to send
    alerts. Until that has been done, it will silently do nothing.

    Examples
    --------
    Specify this class when creating a router. All uncaught exceptions from
    handlers managed by that router will be reported to Slack, if Slack alerts
    are configured.

    .. code-block:: python

       router = APIRouter(route_class=SlackRouteErrorHandler)

    Notes
    -----
    Based on `this StackOverflow question
    <https://stackoverflow.com/questions/61596911/catch-exception-globally-in-fastapi>`__.
    """

    _IGNORED_EXCEPTIONS = (
        HTTPException,
        RequestValidationError,
        StarletteHTTPException,
        SlackIgnoredException,
    )
    """Uncaught exceptions that should not be sent to Slack."""

    _alert_client: ClassVar[SlackWebhookClient | None] = None
    """Global Slack alert client used by `SlackRouteErrorHandler`.

    Initialize with `initialize`. This object caches the alert confguration
    and desired logger for the use of `SlackRouteErrorHandler` as a
    process-global variable, since the route handler only has access to the
    incoming request and global variables.
    """

    @classmethod
    def initialize(
        cls, hook_url: str | SecretStr, application: str, logger: BoundLogger
    ) -> None:
        """Configure Slack alerting.

        Until this function is called, all Slack alerting for uncaught
        exceptions will be disabled.

        Parameters
        ----------
        hook_url
            The URL of the incoming webhook to use to publish the message.
        application
            Name of the application reporting an error.
        logger
            Logger to which to report errors sending messages to Slack.
        """
        cls._alert_client = SlackWebhookClient(hook_url, application, logger)

    @override
    def get_route_handler(
        self,
    ) -> Callable[[Request], Coroutine[Any, Any, Response]]:
        """Wrap route handler with an exception handler."""
        original_route_handler = super().get_route_handler()

        async def wrapped_route_handler(request: Request) -> Response:
            try:
                return await original_route_handler(request)
            except Exception as e:
                # We don't use safir.slack.report_exception here because we
                # want the exception to propagate to the Sentry uncaught
                # exception handler so that we get the full instrumentation
                # from the Sentry FastAPI integration.
                if isinstance(e, self._IGNORED_EXCEPTIONS):
                    raise
                if not self._alert_client:
                    raise
                await self._alert_client.post_exception(e)
                raise

        return wrapped_route_handler



"""Tests for Sentry helpers."""

from typing import override
from unittest.mock import ANY

import pytest
import sentry_sdk
import structlog

from safir.sentry._helpers import report_exception
from safir.slack.blockkit import SentryEventInfo, SlackException
from safir.slack.webhook import SlackIgnoredException, SlackWebhookClient
from safir.testing.sentry import Captured
from safir.testing.slack._mocks import MockSlackWebhook


class SomeError(SlackException):
    """An exception type to test before_send handling."""

    def __init__(self, woo: str, bar: str, user: str | None = None) -> None:
        super().__init__(message="Boom!", user=user)
        self.woo = woo
        self.bar = bar
        self.user = user

    @override
    def to_sentry(self) -> SentryEventInfo:
        info = super().to_sentry()
        info.tags["woo"] = self.woo
        info.contexts["foo"] = {"bar": self.bar}
        return info


def test_env_fingerprint_before_send(
    sentry_fingerprint_items: Captured,
) -> None:
    sentry_sdk.capture_exception(Exception("some error"))
    (error,) = sentry_fingerprint_items.errors
    assert error["fingerprint"] == ["{{ default }}", "some_env"]


def test_sentry_exception_before_send(
    sentry_exception_items: Captured,
) -> None:
    exc = SomeError(woo="hoo", bar="baz", user="someuser")

    sentry_sdk.capture_exception(exc)

    (error,) = sentry_exception_items.errors
    assert error["contexts"]["foo"] == {"bar": "baz"}
    assert error["tags"] == {"woo": "hoo"}
    assert error["user"]["username"] == "someuser"


def test_combined_before_send(sentry_combo_items: Captured) -> None:
    exc = SomeError(woo="hoo", bar="baz", user="someuser")

    sentry_sdk.capture_exception(exc)

    (error,) = sentry_combo_items.errors
    assert error["contexts"]["foo"] == {"bar": "baz"}
    assert error["fingerprint"] == ["{{ default }}", "some_env"]
    assert error["tags"] == {"woo": "hoo"}
    assert error["user"]["username"] == "someuser"


@pytest.mark.asyncio
async def test_report_exception_both(
    mock_slack: MockSlackWebhook, sentry_combo_items: Captured
) -> None:
    logger = structlog.get_logger(__file__)
    client = SlackWebhookClient(mock_slack.url, "App", logger)
    exc = SomeError(woo="hoo", bar="baz", user="someuser")

    await report_exception(exc, slack_client=client)

    assert mock_slack.messages == [
        {
            "blocks": [
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": "Error in App: Boom!",
                        "verbatim": True,
                    },
                },
                {
                    "type": "section",
                    "fields": [
                        {
                            "type": "mrkdwn",
                            "text": "*Exception type*\nSomeError",
                            "verbatim": True,
                        },
                        {
                            "type": "mrkdwn",
                            "text": ANY,
                            "verbatim": True,
                        },
                        {
                            "type": "mrkdwn",
                            "text": "*User*\nsomeuser",
                            "verbatim": True,
                        },
                    ],
                },
                {"type": "divider"},
            ]
        }
    ]

    (error,) = sentry_combo_items.errors
    assert error["contexts"]["foo"] == {"bar": "baz"}
    assert error["fingerprint"] == ["{{ default }}", "some_env"]
    assert error["tags"] == {"woo": "hoo"}
    assert error["user"]["username"] == "someuser"


@pytest.mark.asyncio
async def test_report_exception_sentry_only(
    mock_slack: MockSlackWebhook, sentry_combo_items: Captured
) -> None:
    exc = SomeError(woo="hoo", bar="baz", user="someuser")

    await report_exception(exc, slack_client=None)

    assert mock_slack.messages == []

    (error,) = sentry_combo_items.errors
    assert error["contexts"]["foo"] == {"bar": "baz"}
    assert error["fingerprint"] == ["{{ default }}", "some_env"]
    assert error["tags"] == {"woo": "hoo"}
    assert error["user"]["username"] == "someuser"


@pytest.mark.asyncio
async def test_report_exception_ignores(
    mock_slack: MockSlackWebhook, sentry_combo_items: Captured
) -> None:
    class IgnoredError(SlackIgnoredException):
        pass

    logger = structlog.get_logger(__file__)
    client = SlackWebhookClient(mock_slack.url, "App", logger)
    exc = IgnoredError()

    await report_exception(exc, slack_client=client)

    assert mock_slack.messages == []
    assert sentry_combo_items.errors == []



"""Mocks for captured Sentry information and transport.

Taken from the `Sentry Python SDK's unit tests
<https://github.com/getsentry/sentry-python/blob/c6a89d64db965fe0ece6de10df38ab936af8f5e4/tests/conftest.py>`__.
"""

# Copyright (c) 2018 Functional Software, Inc. dba Sentry
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from dataclasses import dataclass
from typing import Any, override

from sentry_sdk.envelope import Envelope
from sentry_sdk.transport import Transport

__all__ = [
    "Attachment",
    "Captured",
    "TestTransport",
]


@dataclass
class Attachment:
    """Contents and metadata of a Sentry attachment."""

    filename: str
    bytes: bytes
    content_type: str


@dataclass
class Captured:
    """A container for interesting items sent to Sentry."""

    errors: list[dict[str, Any]]
    transactions: list[dict[str, Any]]
    attachments: list[Attachment]


class TestTransport(Transport):
    """A transport that doesn't actually transport anything."""

    def __init__(self) -> None:
        Transport.__init__(self)

    @override
    def capture_envelope(self, envelope: Envelope) -> None:
        """No-op capture_envelope for tests."""



"""Utilities for Pydantic models."""

from ._camel import CamelCaseModel, to_camel_case
from ._types import (
    EnvAsyncPostgresDsn,
    EnvRedisDsn,
    HumanTimedelta,
    IvoaIsoDatetime,
    SecondsTimedelta,
    UtcDatetime,
)
from ._validators import (
    normalize_datetime,
    normalize_isodatetime,
    validate_exactly_one_of,
)

__all__ = [
    "CamelCaseModel",
    "EnvAsyncPostgresDsn",
    "EnvRedisDsn",
    "HumanTimedelta",
    "IvoaIsoDatetime",
    "SecondsTimedelta",
    "UtcDatetime",
    "normalize_datetime",
    "normalize_isodatetime",
    "to_camel_case",
    "validate_exactly_one_of",
]



"""Sentry configuration helpers."""

import os
from typing import Annotated, Any

import sentry_sdk
from pydantic import Field
from pydantic_settings import BaseSettings

from safir.sentry._helpers import before_send_handler

__all__ = ["SentryConfig", "initialize_sentry", "should_enable_sentry"]


class SentryConfig(BaseSettings):
    """Values from environment variables to configure Sentry.

    Most Safir apps should define at least these paramters when initializing
    Sentry. These are taken from enviroment variables and not part of the main
    app config class because we want to initialize Sentry before the app
    configuration is initialized.
    """

    dsn: Annotated[
        str,
        Field(
            validation_alias="SENTRY_DSN",
            description="DSN for sending events to Sentry",
        ),
    ]

    environment: Annotated[
        str,
        Field(
            validation_alias="SENTRY_ENVIRONMENT",
            description=(
                "Phalanx name of the Rubin Science Platform environment"
            ),
        ),
    ]

    traces_sample_rate: Annotated[
        float,
        Field(
            validation_alias="SENTRY_TRACES_SAMPLE_RATE",
            description=(
                "Percentage of transactions to send to Sentry, expressed "
                "as a float between 0 and 1. 0 means send no traces, 1 "
                "means send every trace."
            ),
            ge=0,
            le=1,
        ),
    ] = 0


def should_enable_sentry() -> bool:
    """Return True if we should enable the Sentry integration."""
    return bool(os.environ.get("SENTRY_DSN"))


def initialize_sentry(release: str, **additional_kwargs: Any) -> None:
    """Initialize Sentry with env var values and the safir before_send handler.

    Most Safir apps should provide certain Sentry parameters. This method will
    validate and pass those parameters, as well as any additional parameters,
    to the Sentry SDK init function. It also adds
    `~safir.sentry.before_send_handler` so that
    `~safir.slack.blockkit.SlackException` values can be used.

    Parameters
    ----------
    release
        The version of this application that should be sent \
        with every Sentry event. For most Safir applications, you should pass \
        the value in ``<package>.__version__``.
    """
    if not should_enable_sentry():
        return

    config = SentryConfig()
    kwargs = config.model_dump()
    sentry_sdk.init(
        before_send=before_send_handler,
        release=release,
        **kwargs,
        **additional_kwargs,
    )



"""FastAPI dependencies for the UWS service.

The UWS FastAPI support is initialized by the parent application via this
dependency's ``initialize`` method. It then returns a `UWSFactory` on request
to individual route handlers, which in turn can create other needed objects.
"""

from typing import Annotated, Literal

from fastapi import Depends, Form, Query
from httpx import AsyncClient
from structlog.stdlib import BoundLogger

try:
    from safir.arq import ArqMode, ArqQueue, MockArqQueue, RedisArqQueue
except ImportError as e:
    raise ImportError(
        "The safir.uws module requires the uws extra. "
        "Install it with `pip install safir[uws]`."
    ) from e
from safir.dependencies.http_client import http_client_dependency
from safir.dependencies.logger import logger_dependency

from ._config import UWSConfig
from ._responses import UWSTemplates
from ._results import ResultStore
from ._service import JobService
from ._storage import JobStore

__all__ = [
    "UWSDependency",
    "UWSFactory",
    "create_phase_dependency",
    "runid_post_dependency",
    "uws_dependency",
]


class UWSFactory:
    """Build UWS components.

    Parameters
    ----------
    config
        UWS configuration.
    arq
        arq queue to use.
    result_store
        Signed URL generator for results.
    logger
        Logger to use.
    """

    def __init__(
        self,
        *,
        config: UWSConfig,
        result_store: ResultStore,
        arq: ArqQueue,
        http_client: AsyncClient,
        logger: BoundLogger,
    ) -> None:
        self._config = config
        self._result_store = result_store
        self._arq = arq
        self._http_client = http_client
        self._logger = logger

    def create_result_store(self) -> ResultStore:
        """Return a wrapper around the result storage."""
        return self._result_store

    def create_job_service(self) -> JobService:
        """Create a new UWS job metadata service."""
        return JobService(
            config=self._config,
            arq_queue=self._arq,
            storage=self.create_job_store(),
            logger=self._logger,
        )

    def create_job_store(self) -> JobStore:
        """Create a new UWS job store."""
        return JobStore(self._config, self._http_client)

    def create_templates(self) -> UWSTemplates:
        """Create a new XML renderer for responses."""
        return UWSTemplates()


class UWSDependency:
    """Initializes UWS and provides a UWS factory as a dependency."""

    def __init__(self) -> None:
        self._arq: ArqQueue | None = None
        self._config: UWSConfig
        self._result_store: ResultStore

    async def __call__(
        self,
        http_client: Annotated[AsyncClient, Depends(http_client_dependency)],
        logger: Annotated[BoundLogger, Depends(logger_dependency)],
    ) -> UWSFactory:
        if not self._arq:
            raise RuntimeError("UWSDependency not initialized")
        return UWSFactory(
            config=self._config,
            result_store=self._result_store,
            arq=self._arq,
            http_client=http_client,
            logger=logger,
        )

    async def initialize(self, config: UWSConfig) -> None:
        """Initialize the UWS subsystem.

        Parameters
        ----------
        config
            UWS configuration.
        """
        self._config = config
        self._result_store = ResultStore(config)
        if not self._arq:
            if config.arq_mode == ArqMode.production:
                settings = config.arq_redis_settings
                self._arq = await RedisArqQueue.initialize(settings)
            else:
                self._arq = MockArqQueue()

    def override_arq_queue(self, arq_queue: ArqQueue) -> None:
        """Change the arq used in subsequent invocations.

        This method is probably only useful for the test suite.

        Parameters
        ----------
        arq_queue
            New arq queue.
        """
        self._arq = arq_queue


uws_dependency = UWSDependency()


async def create_phase_dependency(
    *,
    get_phase: Annotated[
        Literal["RUN"] | None,
        Query(title="Immediately start job", alias="phase"),
    ] = None,
    post_phase: Annotated[
        Literal["RUN"] | None,
        Form(title="Immediately start job", alias="phase"),
    ] = None,
) -> Literal["RUN"] | None:
    """Parse the optional phase parameter to an async job creation.

    Allow ``phase=RUN`` to be specified in either the query or the POST
    parameters, which says that the job should be immediately started.
    """
    return post_phase or get_phase


async def runid_post_dependency(
    *,
    runid: Annotated[
        str | None,
        Form(
            title="Run ID for job",
            description=(
                "An opaque string that is returned in the job metadata and"
                " job listings. May be used by the client to associate jobs"
                " with specific larger operations."
            ),
        ),
    ] = None,
) -> str | None:
    """Parse the run ID from POST parameters."""
    return runid



"""Handlers for the UWS API to a service.

The handlers defined in ``uws_router`` should be reusable for any IVOA service
that implements UWS. The other functions dynamically define additional
handlers that use dependencies provided by the application, so that the
application can define the parameters to a job.
"""

from datetime import datetime, timedelta
from typing import Annotated, Literal

from fastapi import APIRouter, Depends, Form, Query, Request, Response
from fastapi.responses import PlainTextResponse, RedirectResponse

try:
    from vo_models.uws import Jobs, JobSummary, Results
    from vo_models.uws.types import ExecutionPhase
except ImportError as e:
    raise ImportError(
        "The safir.uws module requires the uws extra. "
        "Install it with `pip install safir[uws]`."
    ) from e

from safir.datetime import isodatetime
from safir.dependencies.gafaelfawr import (
    auth_delegated_token_dependency,
    auth_dependency,
)
from safir.pydantic import IvoaIsoDatetime
from safir.slack.webhook import SlackRouteErrorHandler

from ._config import UWSRoute
from ._dependencies import (
    UWSFactory,
    create_phase_dependency,
    runid_post_dependency,
    uws_dependency,
)
from ._exceptions import DataMissingError
from ._models import ParametersModel

uws_router = APIRouter(route_class=SlackRouteErrorHandler)
"""FastAPI router for all external handlers."""

__all__ = [
    "install_async_post_handler",
    "install_availability_handler",
    "install_sync_get_handler",
    "install_sync_post_handler",
    "uws_router",
]


@uws_router.get(
    "",
    description=(
        "List all existing jobs for the current user. Jobs will be sorted"
        " by creation date, with the most recently created listed first."
    ),
    responses={
        200: {
            "content": {
                "application/xml": {"schema": Jobs.model_json_schema()}
            }
        }
    },
    summary="Async job list",
)
async def get_job_list(
    *,
    request: Request,
    phase: Annotated[
        list[ExecutionPhase] | None,
        Query(
            title="Execution phase",
            description="Limit results to the provided execution phases",
        ),
    ] = None,
    after: Annotated[
        datetime | None,
        Query(
            title="Creation date",
            description="Limit results to jobs created after this date",
        ),
    ] = None,
    last: Annotated[
        int | None,
        Query(
            title="Number of jobs",
            description="Return at most the given number of jobs",
        ),
    ] = None,
    token: Annotated[str, Depends(auth_delegated_token_dependency)],
    uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
) -> Response:
    job_service = uws_factory.create_job_service()
    jobs = await job_service.list_jobs(
        token,
        base_url=str(request.url_for("get_job_list")),
        phases=phase,
        after=after,
        count=last,
    )
    xml = jobs.to_xml(skip_empty=True)
    return Response(content=xml, media_type="application/xml")


@uws_router.get(
    "/{job_id}",
    responses={
        200: {
            "content": {
                "application/xml": {"schema": JobSummary.model_json_schema()}
            }
        }
    },
    summary="Job details",
)
async def get_job(
    job_id: str,
    *,
    request: Request,
    wait: Annotated[
        int | None,
        Query(
            title="Wait for status changes",
            description=(
                "Maximum number of seconds to wait or -1 to wait for as long"
                " as the server permits"
            ),
        ),
    ] = None,
    phase: Annotated[
        ExecutionPhase | None,
        Query(
            title="Initial phase for waiting",
            description=(
                "When waiting for status changes, consider this to be the"
                " initial execution phase. If the phase has already changed,"
                " return immediately. This parameter should always be"
                " provided when wait is used."
            ),
        ),
    ] = None,
    token: Annotated[str, Depends(auth_delegated_token_dependency)],
    uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
) -> Response:
    job_service = uws_factory.create_job_service()
    result_store = uws_factory.create_result_store()
    job = await job_service.get_summary(
        token, job_id, signer=result_store, wait_seconds=wait, wait_phase=phase
    )
    xml = job.to_xml(skip_empty=True)
    return Response(content=xml, media_type="application/xml")


@uws_router.delete(
    "/{job_id}",
    status_code=303,
    response_class=RedirectResponse,
    summary="Delete a job",
)
async def delete_job(
    job_id: str,
    *,
    request: Request,
    token: Annotated[str, Depends(auth_delegated_token_dependency)],
    uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
) -> str:
    job_service = uws_factory.create_job_service()
    await job_service.delete(token, job_id)
    return str(request.url_for("get_job_list"))


@uws_router.post(
    "/{job_id}",
    description=(
        "Alternate job deletion mechanism for clients that cannot use DELETE."
    ),
    response_class=RedirectResponse,
    status_code=303,
    summary="Delete a job",
)
async def delete_job_via_post(
    job_id: str,
    *,
    action: Annotated[
        Literal["DELETE"] | None,
        Form(
            title="Action to perform",
            description="Mandatory, must be set to DELETE",
        ),
    ] = None,
    request: Request,
    token: Annotated[str, Depends(auth_delegated_token_dependency)],
    uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
) -> str:
    job_service = uws_factory.create_job_service()
    await job_service.delete(token, job_id)
    return str(request.url_for("get_job_list"))


@uws_router.get(
    "/{job_id}/destruction",
    response_class=PlainTextResponse,
    summary="Destruction time for job",
)
async def get_job_destruction(
    job_id: str,
    *,
    token: Annotated[str, Depends(auth_delegated_token_dependency)],
    uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
) -> str:
    job_service = uws_factory.create_job_service()
    job = await job_service.get(token, job_id)
    return isodatetime(job.destruction_time)


@uws_router.post(
    "/{job_id}/destruction",
    response_class=RedirectResponse,
    status_code=303,
    summary="Change job destruction time",
)
async def post_job_destruction(
    job_id: str,
    *,
    destruction: Annotated[
        IvoaIsoDatetime,
        Form(
            title="New destruction time",
            description="Must be in ISO 8601 format.",
            examples=["2021-09-10T10:01:02Z"],
        ),
    ],
    request: Request,
    token: Annotated[str, Depends(auth_delegated_token_dependency)],
    uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
) -> str:
    job_service = uws_factory.create_job_service()
    await job_service.update_destruction(token, job_id, destruction)
    return str(request.url_for("get_job", job_id=job_id))


@uws_router.get(
    "/{job_id}/error",
    responses={
        200: {"content": {"application/x-votable+xml": {}}},
        404: {"description": "Job not found or job did not fail"},
    },
    summary="Job error",
)
async def get_job_error(
    job_id: str,
    *,
    request: Request,
    token: Annotated[str, Depends(auth_delegated_token_dependency)],
    uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
) -> Response:
    job_service = uws_factory.create_job_service()
    errors = await job_service.get_error(token, job_id)
    if not errors:
        raise DataMissingError(f"Job {job_id} did not fail")
    templates = uws_factory.create_templates()

    # Wobbly supports storing multiple errors for a job, but currently the
    # templating into a VOTable only supports a single error and the code to
    # store the error only stores a single error. For now, return only the
    # first element of the list of errors.
    return templates.error(request, errors[0])


@uws_router.get(
    "/{job_id}/executionduration",
    response_class=PlainTextResponse,
    summary="Execution duration of job",
)
async def get_job_execution_duration(
    job_id: str,
    *,
    token: Annotated[str, Depends(auth_delegated_token_dependency)],
    uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
) -> str:
    job_service = uws_factory.create_job_service()
    job = await job_service.get(token, job_id)
    if job.execution_duration:
        return str(int(job.execution_duration.total_seconds()))
    else:
        return "0"


@uws_router.post(
    "/{job_id}/executionduration",
    response_class=RedirectResponse,
    status_code=303,
    summary="Change job execution duration",
)
async def post_job_execution_duration(
    job_id: str,
    *,
    executionduration: Annotated[
        int,
        Form(
            title="New execution duration",
            description="Integer seconds of wall clock time.",
            examples=[14400],
            ge=0,
        ),
    ],
    request: Request,
    token: Annotated[str, Depends(auth_delegated_token_dependency)],
    uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
) -> str:
    job_service = uws_factory.create_job_service()
    duration = None
    if executionduration > 0:
        duration = timedelta(seconds=executionduration)
    await job_service.update_execution_duration(token, job_id, duration)
    return str(request.url_for("get_job", job_id=job_id))


@uws_router.get(
    "/{job_id}/owner",
    response_class=PlainTextResponse,
    summary="Owner of job",
)
async def get_job_owner(
    job_id: str,
    *,
    token: Annotated[str, Depends(auth_delegated_token_dependency)],
    uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
) -> str:
    job_service = uws_factory.create_job_service()
    job = await job_service.get(token, job_id)
    return job.owner


@uws_router.get(
    "/{job_id}/parameters",
    responses={200: {"content": {"application/xml": {}}}},
    summary="Job parameters",
)
async def get_job_parameters(
    job_id: str,
    *,
    request: Request,
    token: Annotated[str, Depends(auth_delegated_token_dependency)],
    uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
) -> Response:
    job_service = uws_factory.create_job_service()
    job = await job_service.get_summary(token, job_id)
    if not job.parameters:
        raise DataMissingError(f"Job {job_id} has no parameters")
    xml = job.parameters.to_xml(skip_empty=True)
    return Response(content=xml, media_type="application/xml")


@uws_router.get(
    "/{job_id}/phase",
    response_class=PlainTextResponse,
    summary="Phase of job",
)
async def get_job_phase(
    job_id: str,
    *,
    token: Annotated[str, Depends(auth_delegated_token_dependency)],
    uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
) -> str:
    job_service = uws_factory.create_job_service()
    job = await job_service.get(token, job_id)
    return job.phase.value


@uws_router.post(
    "/{job_id}/phase",
    response_class=RedirectResponse,
    status_code=303,
    summary="Start or abort job",
)
async def post_job_phase(
    job_id: str,
    *,
    phase: Annotated[
        Literal["RUN", "ABORT"] | None,
        Form(
            title="Job state change",
            summary="RUN to start the job, ABORT to abort the job.",
        ),
    ] = None,
    request: Request,
    user: Annotated[str, Depends(auth_dependency)],
    token: Annotated[str, Depends(auth_delegated_token_dependency)],
    uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
) -> str:
    job_service = uws_factory.create_job_service()
    if phase == "ABORT":
        await job_service.abort(token, job_id)
    elif phase == "RUN":
        await job_service.start(token, user, job_id)
    return str(request.url_for("get_job", job_id=job_id))


@uws_router.get(
    "/{job_id}/quote",
    response_class=PlainTextResponse,
    summary="Quote for job",
)
async def get_job_quote(
    job_id: str,
    *,
    token: Annotated[str, Depends(auth_delegated_token_dependency)],
    uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
) -> str:
    job_service = uws_factory.create_job_service()
    job = await job_service.get(token, job_id)
    if job.quote:
        return isodatetime(job.quote)
    else:
        # The UWS standard says to return an empty text/plain response in this
        # case, as weird as it might look.
        return ""


@uws_router.get(
    "/{job_id}/results",
    responses={
        200: {
            "content": {
                "application/xml": {"schema": Results.model_json_schema()}
            }
        }
    },
    summary="Job results",
)
async def get_job_results(
    job_id: str,
    *,
    request: Request,
    token: Annotated[str, Depends(auth_delegated_token_dependency)],
    uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
) -> Response:
    job_service = uws_factory.create_job_service()
    result_store = uws_factory.create_result_store()
    job = await job_service.get_summary(token, job_id, signer=result_store)
    if not job.results:
        raise DataMissingError(f"Job {job_id} has no results")
    xml = job.results.to_xml(skip_empty=True)
    return Response(content=xml, media_type="application/xml")


def install_availability_handler(router: APIRouter) -> None:
    """Construct a default handler for the VOSI ``/availability`` interface.

    Parameters
    ----------
    router
        Router into which to install the handler.
    """

    @router.get(
        "/availability",
        description="VOSI-availability resource for the service",
        responses={200: {"content": {"application/xml": {}}}},
        summary="IVOA service availability",
    )
    async def get_availability(
        request: Request,
        uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
    ) -> Response:
        job_service = uws_factory.create_job_service()
        availability = await job_service.availability()
        xml = availability.to_xml(skip_empty=True)
        return Response(content=xml, media_type="application/xml")


def install_async_post_handler(router: APIRouter, route: UWSRoute) -> None:
    """Construct the POST handler for creating an async job.

    Parameters
    ----------
    router
        Router into which to install the handler.
    route
        Configuration for this route.
    """

    @router.post(
        "/jobs",
        response_class=RedirectResponse,
        status_code=303,
        summary=route.summary,
        description=route.description,
    )
    async def create_job(
        *,
        request: Request,
        phase: Annotated[
            Literal["RUN"] | None, Depends(create_phase_dependency)
        ] = None,
        parameters: Annotated[ParametersModel, Depends(route.dependency)],
        runid: Annotated[str | None, Depends(runid_post_dependency)],
        user: Annotated[str, Depends(auth_dependency)],
        token: Annotated[str, Depends(auth_delegated_token_dependency)],
        uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
    ) -> str:
        job_service = uws_factory.create_job_service()
        job = await job_service.create(token, parameters, run_id=runid)
        if phase == "RUN":
            await job_service.start(token, user, job.id)
        return str(request.url_for("get_job", job_id=job.id))


def install_sync_post_handler(router: APIRouter, route: UWSRoute) -> None:
    """Construct the POST handler for creating a sync job.

    Parameters
    ----------
    router
        Router into which to install the handler.
    route
        Configuration for this route.
    """

    @router.post(
        "/sync",
        response_class=RedirectResponse,
        status_code=303,
        summary=route.summary,
        description=route.description,
    )
    async def post_sync(
        *,
        runid: Annotated[str | None, Depends(runid_post_dependency)],
        parameters: Annotated[ParametersModel, Depends(route.dependency)],
        user: Annotated[str, Depends(auth_dependency)],
        token: Annotated[str, Depends(auth_delegated_token_dependency)],
        uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
    ) -> str:
        job_service = uws_factory.create_job_service()
        result_store = uws_factory.create_result_store()
        result = await job_service.run_sync(
            token, user, parameters, runid=runid
        )
        return result_store.sign_url(result).url


def install_sync_get_handler(router: APIRouter, route: UWSRoute) -> None:
    """Construct the GET handler for creating a sync job.

    Parameters
    ----------
    router
        Router into which to install the handler.
    route
        Configuration for this route.
    """

    @router.get(
        "/sync",
        response_class=RedirectResponse,
        status_code=303,
        summary=route.summary,
        description=route.description,
    )
    async def get_sync(
        *,
        runid: Annotated[
            str | None,
            Query(
                title="Run ID for job",
                description=(
                    "An opaque string that is returned in the job metadata and"
                    " job listings. May be used by the client to associate"
                    " jobs with specific larger operations."
                ),
            ),
        ] = None,
        parameters: Annotated[ParametersModel, Depends(route.dependency)],
        user: Annotated[str, Depends(auth_dependency)],
        token: Annotated[str, Depends(auth_delegated_token_dependency)],
        uws_factory: Annotated[UWSFactory, Depends(uws_dependency)],
    ) -> str:
        job_service = uws_factory.create_job_service()
        result_store = uws_factory.create_result_store()
        result = await job_service.run_sync(
            token, user, parameters, runid=runid
        )
        return result_store.sign_url(result).url



"""Validation functions for Pydantic models."""

from __future__ import annotations

from collections.abc import Callable
from datetime import UTC, datetime
from typing import Any

from pydantic import BaseModel

from safir.datetime import parse_isodatetime

__all__ = [
    "normalize_datetime",
    "normalize_isodatetime",
    "validate_exactly_one_of",
]


def normalize_datetime(v: Any) -> datetime | None:
    """Pydantic field validator for datetime fields.

    Supports `~datetime.datetime` fields given as either datetime objects or
    seconds since epoch (not the other types Pydantic natively supports) and
    ensures that the resulting datetime object is timezone-aware and in the
    UTC timezone.

    Parameters
    ----------
    v
        Field representing a `~datetime.datetime`.

    Returns
    -------
    datetime.datetime or None
        The timezone-aware `~datetime.datetime` or `None` if the input was
        `None`.

    Raises
    ------
    ValueError
        Raised if the input could not be parsed as a `~datetime.datetime`.

    Notes
    -----
    Prefer to use the `~safir.pydantic.UtcDatetime` type instead of using this
    function as a validator.

    Examples
    --------
    Here is a partial model that uses this function as a field validator.

    .. code-block:: python

       class Info(BaseModel):
           last_used: datetime | None = Field(
               None,
               title="Last used",
               description="When last used in seconds since epoch",
               examples=[1614986130],
           )

           _normalize_last_used = field_validator("last_used", mode="before")(
               normalize_datetime
           )
    """
    if v is None:
        return v
    elif isinstance(v, int):
        return datetime.fromtimestamp(v, tz=UTC)
    elif not isinstance(v, datetime):
        raise ValueError("Must be a datetime or seconds since epoch")
    elif v.tzinfo and v.tzinfo.utcoffset(v) is not None:
        return v.astimezone(UTC)
    else:
        return v.replace(tzinfo=UTC)


def normalize_isodatetime(v: Any) -> datetime | None:
    """Pydantic field validator for datetime fields in ISO format.

    This field validator requires a subset of the ISO 8601 date and time
    format, ``YYYY-MM-DD[THH:MM:SS[.mmm]][Z]``. Regardless of whether the
    trailing ``Z`` is included, the date and time are interpreted as being in
    UTC, not local time. This format is compatible with Kubernetes, the IVOA
    DALI standard, and the format produced by `safir.datetime.isodatetime`.

    It should be used when the other formats supported by Pydantic by default
    (such as dates and times in other timezones) shouldn't be allowed, such as
    when strict conformance with the IVOA standard is desired.

    Parameters
    ----------
    v
        Field representing a `~datetime.datetime`.

    Returns
    -------
    datetime.datetime or None
        The timezone-aware `~datetime.datetime` or `None` if the input was
        `None`.

    Raises
    ------
    ValueError
        Raised if the provided time string is not in the correct format.

    Notes
    -----
    Prefer to use the `~safir.pydantic.IvoaIsoDatetime` type instead of using
    this function as a validator.

    Examples
    --------
    Here is a partial model that uses this function as a field validator.

    .. code-block:: python

       class Info(BaseModel):
           last_used: datetime | None = Field(
               None,
               title="Last used",
               description="Date and time last used",
               examples=["2023-01-25T15:44:34Z"],
           )

           _normalize_last_used = field_validator("last_used", mode="before")(
               normalize_isodatetime
           )
    """
    if v is None:
        return None
    if not isinstance(v, str):
        msg = "Must be a string in YYYY-MM-DD[THH:MM:SS[:mmm]][Z] format"
        raise ValueError(msg)
    return parse_isodatetime(v)


def validate_exactly_one_of(
    *settings: str,
) -> Callable[[BaseModel], BaseModel]:
    """Generate a model validator imposing a one and only one constraint.

    Sometimes, models have a set of attributes of which one and only one may
    be set. Ideally this is represented properly in the type system, but
    occasionally it's more convenient to use a model validator. This is a
    model validator generator that can produce a model validator function that
    ensures one and only one of an arbitrary set of attributes must be set.

    Parameters
    ----------
    *settings
        List of names of attributes, of which one and only one must be set.
        At least two attribute names must be listed.

    Returns
    -------
    Callable
        Resulting model validator.

    Examples
    --------
    Use this inside a Pydantic class as a model validator as follows:

    .. code-block:: python

       class Foo(BaseModel):
           foo: Optional[str] = None
           bar: Optional[str] = None
           baz: Optional[str] = None

           _validate_options = model_validator(mode="after")(
               validate_exactly_one_of("foo", "bar", "baz")
           )

    The attribute listed as the first argument to the ``model_validator`` call
    must be the last attribute in the model definition so that any other
    attributes have already been seen.
    """
    if len(settings) < 2:
        msg = "validate_exactly_one_of takes at least two field names"
        raise ValueError(msg)

    if len(settings) == 2:
        options = f"{settings[0]} and {settings[1]}"
    else:
        options = ", ".join(settings[:-1]) + ", and " + settings[-1]

    def validator[T: BaseModel](model: T) -> T:
        seen = False
        for setting in settings:
            if getattr(model, setting, None) is not None:
                if seen:
                    raise ValueError(f"only one of {options} may be given")
                seen = True
        if not seen:
            raise ValueError(f"one of {options} must be given")
        return model

    return validator



"""Redis database support."""

from ._storage import (
    DeserializeError,
    EncryptedPydanticRedisStorage,
    PydanticRedisStorage,
)

__all__ = [
    "DeserializeError",
    "EncryptedPydanticRedisStorage",
    "PydanticRedisStorage",
]
